import warnings

def skip_line(l):
    """Return if the line should be skipped (empty lines or comment only)"""
    x = l.strip()
    return (x.startswith('!') or x.startswith('*') or x.startswith('read rtf') or x == '')

def comment_parser(line):
    """Parse and separate data fields and comments"""
    if line.find('!') == -1:
        # line has no comments
        return line, None
    # return fields string, comments
    line, desc = line.split('!', maxsplit = 1)[:2]
    return line, desc.strip().lower()

def mass_parser(line):
    """Parse lines with keyword MASS"""
    field_str, desc = comment_parser(line)
    # fields: [key, deprecated_entry, atom_type, atomic_mass]
    fields = field_str.split()
    if len(fields) == 4:
        key, _, atom_type, mass = fields
        element = None
    elif len(fields) == 5:
        key, _, atom_type, mass, element = fields
    return atom_type, mass, desc, element

def decl_parser(line):
    """Parse line with keyword DECLare"""
    field_str = comment_parser(line)[0]
    atom = field_str.split()[-1]
    return atom

def defa_parser(line):
    """Parse default terminal patches"""
    field_str = comment_parser(line)[0]
    fields = field_str.split()
    # find where these two keyword located in the list
    first_key = 'FIRS' if 'FIRS' in fields else 'FIRST'
    last_key = 'LAST'
    first_i = fields.index(first_key)
    last_i = fields.index(last_key)
    first_patch = fields[first_i+1]
    last_patch = fields[last_i+1]
    return first_patch, last_patch

def auto_parser(line):
    """Parse keyword AUTOGEN"""
    field_str = comment_parser(line)[0]
    autogen_ops = field_str.split()[1:]
    return autogen_ops

def resi_parser(line):
    """Parse the keyword RESI for a new residue"""
    field_str, desc = comment_parser(line)
    # [key, residue_name, total_charge]
    key, res_name, total_charge = field_str.split()[:3]
    return res_name, total_charge, desc

def atom_parser(line):
    """Parse the keyword ATOM"""
    field_str = comment_parser(line)[0]
    key, atom_name, atom_type, charge = field_str.split()
    return atom_name, atom_type, charge

def pairwise_parser(line):
    """Parser for lines with pairwise atoms separated by space"""
    field_str = comment_parser(line)[0]
    # The first field is the keyword
    # followed by atom type in pairs
    fields = field_str.split()[1:]
    # combine each pair into list of tuples
    if len(fields) % 2 == 1:
        raise ValueError(
            f'Odd number of atoms! Cannot group into pairs: {line}'
        )
    atom_pairs = list(zip(fields[::2], fields[1::2]))
    return atom_pairs

def quad_parser(line):
    """Parse IMPR keyword for group of 4 atoms"""
    field_str = comment_parser(line)[0]
    fields = field_str.split()[1:]
    # combine each pair into list of tuples
    if len(fields) % 4 != 0:
        raise ValueError(
            f'Invalid length of topology specification: {line}\n'
            'Multiples  of 4 atoms required'
        )
    ##TODO: use for loop to combine into list of tuples
    quads = []
    for i in range(0, len(fields), 4):
        quads.append(tuple(fields[i:i+4]))
    # quad = (tuple(fields[:4]), tuple(fields[4:]))
    return quads

def octa_parser(line):
    """Parse CMAP keyword for a pair of 4 atoms"""
    field_str = comment_parser(line)[0]
    fields = field_str.split()[1:]
    if len(fields) != 8:
        raise ValueError(
            f'Invalid length of topology specification: {line}\n'
            '8 atoms required'
        )
    # combine into tuple of tuples with shape (2,4)
    octa = (tuple(fields[:4]), tuple(fields[4:]))
    return octa

def ic_parser(line):
    """
    Parse internal coordinate entries. 9 fields total

    fields: [key, I,J,K,L, R(I(J/K)), T(I(JK/KJ)), PHI, T(JKL), R(KL)]
    IJKL are atoms, star (*) on atom K indicate that it is an improper structure
    Strucutres specified for both chain and improper (branch). If any value is
    zero, it is not specified, and the value should be obtained from the parameter
    file.

        R(I(J/K)): dist of I-J/I-K
        T(I(JK/KJ)): angle of I-J-K/I-K-J
        PHI: Dihedral
        T(JKL): angel of J-K-L
        R(KL): dist of K-L
    """
    field_str = comment_parser(line)[0]
    fields = field_str.split()[1:]
    if len(fields) != 9:
        raise ValueError(
            f'Invalid length of topology specification: {line}\n'
            '9 fields required'
        )
    i, j, k, l, r_ij, t_ijk, phi, t_jkl, r_kl = fields
    if k.startswith('*'):
        is_improper = True
        first_angle = 'R(I-K)'
        dihe = 'T(I-K-J)'
    else:
        is_improper = False
        first_angle = 'R(I-J)'
        dihe = 'T(I-J-K)'

    key =  (i, j, k.lstrip('*'), l)
    key = tuple(s.upper() for s in key)
    entry = {
        'improper': is_improper,
        # 'I': i, 'J': j, 'K': k, 'L': l,
        first_angle: float(r_ij), 
        dihe: float(t_ijk),
        'Phi': float(phi),
        'T(J-K-L)': float(t_jkl),
        'R(K-L)': float(r_kl)
    }
    for k, v in entry.items():
        if k not in ('improper', 'Phi') and v == 0:
            entry[k] = None
    return key, entry

##TODO: delete parser need to handle multiple values other than atom type
def delete_parser(line):
    """
    Parse the keyword DELETE. Keyword has to be followed by the type of the 
    data to be deleted and the value
    """
    field_str = comment_parser(line)[0]
    fields = field_str.split()[1:]
    if len(fields) < 2:
        raise ValueError(
            f'Invalid length of topology specification: {line}\n'
            'at least 2 fields required: key, data'
        )
    return (fields[0],fields[1])

class RTFParser:
    """A parser class to load rtf (residue topology files) into dictionary.
    Parser is initialized with RTF file from file path. If any lines from the 
    file are not parsed, they will be stored in the unparsed_lines"""
    def __init__(self, file_path=None, rtf_block=None):
        self.rtf_version = None
        self.topo_dict = None
        self.decl_peptide_atoms = []
        self.default_patchs = {'FIRST':None,'LAST':None}
        self.default_autogen = None
        self.unparsed_lines = []
        self.mass_dict = {}
        if rtf_block is not None:
            self.lines = [l.strip() for l in rtf_block.split('\n') if not skip_line(l)]
        elif file_path is not None:
            with open(file_path, 'r', encoding='utf-8') as f:
                self.lines = [l.strip() for l in f.readlines() if not skip_line(l)]
        else:
            raise ValueError('Either file_path or rtf_block has to be provided!')
        self._parse_lines()
        n_unparsed = len(self.unparsed_lines)
        if n_unparsed > 0:
            warnings.warn(
                f"Failed to parse {n_unparsed} lines from {file_path}"
            )

    def _parse_lines(self):
        ##TODO: Need to handle keywords DIHE, ANGLE, and PATCH
        self.rtf_version = '.'.join(self.lines[0].strip().split())
        self.topo_dict = {}
        for l in self.lines[1:]:
            l = l.upper()
            if l.startswith('MASS'):
                symbol, mass, desc, element = mass_parser(l)
                self.mass_dict.update({symbol: (float(mass), desc, element)})
            elif l.startswith('DECL'):
                atom = decl_parser(l)
                self.decl_peptide_atoms.append(atom)
            elif l.startswith('DEFA'):
                first_patch, last_patch = defa_parser(l)
                self.default_patchs['FIRST'] = first_patch
                self.default_patchs['LAST'] = last_patch
            elif l.startswith('AUTO'):
                self.default_autogen = auto_parser(l)
            elif l.startswith('RESI') or l.startswith('PRES'):
                res_name, total_charge, desc = resi_parser(l)
                if res_name not in self.topo_dict:
                    # index first group
                    cur_group_i = -1
                    cur_res = {
                        'desc': desc,
                        'total_charge': float(total_charge),
                        'atoms':{
                        },
                        'bonds':{
                            'single':[],
                            'double':[],
                            'triple':[],
                            'aromatic':[]
                        },
                        'impropers':[],
                        'cmap':[],
                        'ic':{},
                        'is_patch': l.startswith('PRES')
                    }
                self.topo_dict.update({res_name: cur_res})
            elif l.startswith('GROU'):
                # Update group number
                cur_group_i += 1
                cur_atom_group = {cur_group_i: {}}
                cur_res['atoms'].update(cur_atom_group)
            elif l.startswith('ATOM'):
                if cur_group_i == -1:
                    # if no GROUP keyword exist for patch, create a single group
                    cur_group_i = 0
                    cur_atom_group = {cur_group_i: {}}
                    cur_res['atoms'].update(cur_atom_group)
                atom_name, atom_type, atom_charge = atom_parser(l)
                if atom_type not in self.mass_dict:
                    # CGENFF rtf files do not have MASS table
                    cur_atom_dict = {
                        atom_name:
                        {
                            'atom_type': atom_type, 
                            'charge': float(atom_charge), 
                        }
                    }
                else:
                    cur_atom_dict = {
                        atom_name: 
                        {
                            'atom_type': atom_type, 
                            'charge': float(atom_charge), 
                            'mass': self.mass_dict[atom_type][0],
                            'desc': self.mass_dict[atom_type][1],
                            'element': self.mass_dict[atom_type][2]
                        }
                    }
                cur_atom_group[cur_group_i].update(cur_atom_dict)
            elif l.startswith('DONO'):
                if 'H_donors' not in cur_res:
                    cur_res['H_donors'] = []
                donors = tuple(l.split()[1:])
                cur_res['H_donors'].append(donors)
            elif l.startswith('ACCE'):
                if 'H_acceptors' not in cur_res:
                    cur_res['H_acceptors'] = []
                acceptors = tuple(l.split()[1:])
                cur_res['H_acceptors'].append(acceptors)
            elif l.startswith('BOND'):
                single_bonds = pairwise_parser(l)
                cur_res['bonds']['single'].extend(single_bonds)
            elif l.startswith('DOUB'):
                double_bonds = pairwise_parser(l)
                cur_res['bonds']['double'].extend(double_bonds)
            elif l.startswith('TRIP'):
                triple_bonds = pairwise_parser(l)
                cur_res['bonds']['triple'].extend(triple_bonds)
            elif l.startswith('AROM'):
                aromatic_bonds = pairwise_parser(l)
                cur_res['bonds']['triple'].extend(aromatic_bonds)
            elif l.startswith('IMPR'):
                # Improper (branching structures)
                ##TODO: quad_parser is not correct for IMPR
                impropers = quad_parser(l)
                cur_res['impropers'].extend(impropers)
            elif l.startswith('CMAP'):
                # Dihedral crossterm energy correction map
                cmap = octa_parser(l)
                cur_res['cmap'].append(cmap)
            elif (
                l.startswith('IC') or l.startswith('BILD')
            ):
                # Internal Coordinates
                ic_key, ic_param_dict = ic_parser(l)
                cur_res['ic'][ic_key] = ic_param_dict
            elif l.startswith('DELE'):
                if 'delete' not in cur_res:
                    cur_res['delete'] = []
                delete_entry = delete_parser(l)
                cur_res['delete'].append(delete_entry)
            elif l.startswith('DIHE'):
                pass
            elif l.startswith('ANGL'):
                pass
            elif l.startswith('PATC'):
                pass
            elif l.startswith('END'):
                break
            else:
                self.unparsed_lines.append(l)



import pycharmm
from pycharmm import write as pcm_write
from pycharmm import NonBondedScript
from pycharmm import crystal, image

from crimm.Adaptors.pyCHARMMAdaptors import (
    load_chain, load_topology, load_water, load_ions, load_ligands,
    create_water_hs_from_charmm, fetch_coords_from_charmm, sd_minimize,
    patch_disu_from_model
)

from crimm import fetch_rcsb
from crimm.Modeller.Solvator import Solvator
from crimm.Modeller.LoopBuilder import ChainLoopBuilder

from crimm.Modeller.CoordManipulator import CoordManipulator
from crimm.Modeller import TopologyGenerator
from crimm.Adaptors.PropKaAdaptors import PropKaProtonator


CGENFF_PATH = "/export/app/cgenff/silcsbio.2024.1/cgenff/cgenff"
PDBIDS = ['2HZI', '5IGV', '1BG8', '3Q4K', '4PTI', '5IEV']
PDBID = '5IEV'
model = fetch_rcsb(
    PDBID,
    include_solvent=True,
    use_bio_assembly=True,
    organize=True
)

## Place the model center to (0, 0, 0) and place the principle axis along x-axis
coord_man = CoordManipulator()
coord_man.load_entity(model)
coord_man.orient_coords()

# build missing loops if exist
for chain in model.protein:
    if not chain.is_continuous():
        # chain can be built in place now by specifying `inplace = True`
        looper = ChainLoopBuilder(chain, inplace = True)
        # missing terminals will also be built if `include_terminal = True`
        looper.build_from_alphafold(include_terminal = False)

topo = TopologyGenerator(CGENFF_PATH)
topo.generate_model(model, coerce=True)

protonator = PropKaProtonator(topo, pH = 7.4)
protonator.load_model(model)
protonator.apply_patches()

load_topology(model.topology_loader)
for chain in model.protein:
    load_chain(chain)

# we need to patch disulfide bonds in CHARMM
# the disu info is stored in model under model.connect_dict
patch_disu_from_model(model)

load_ligands(model.ligand+model.phos_ligand+model.co_solvent)

## Minimize the protein and ligand first
# Specify nonbonded python object called my_nbonds - this just sets it up
# equivalant CHARMM scripting command: 
# nbonds cutnb 18 ctonnb 13 ctofnb 17 cdie eps 1 atom vatom fswitch vfswitch
non_bonded_script = NonBondedScript(
    cutnb=18.0, ctonnb=13.0, ctofnb=17.0,
    eps=1.0,
    cdie=True,
    atom=True, vatom=True,
    fswitch=True, vfswitch=True
)
# select the C-alpha atoms for harmonic restraints
cons_harm_atoms = pycharmm.SelectAtoms(atom_type='CA')
ener_dict = sd_minimize(
    300, non_bonded_script, cons_harm_selection=cons_harm_atoms
)

## Update the coordinates of the model from CHARMM
fetch_coords_from_charmm(
    model.protein+model.ligand+model.co_solvent+model.phos_ligand
)

solvator = Solvator(model)
solvator.solvate(
    cutoff=8.0, solvcut=2.1, remove_existing_water=False, orient_coords=False
)
solvator.add_balancing_ions()

load_water(model.solvent)
load_ions(model.ion)
create_water_hs_from_charmm(model)

## Setup PBC and Minimize Water
# organize segids and ion types for image and cons_fix
non_solvent_segids = set()
all_ion_types = set()
for chain in model:
    if chain.chain_type == 'Solvent':
        continue
    elif chain.chain_type == 'Ion':
        for res in chain:
            all_ion_types.add(res.resname)
    else:
        for res in chain:
            non_solvent_segids.add(res.segid)

# CHARMM scripting: crystal define cubic @boxsize @boxsize @boxsize 90 90 90
crystal.define_cubic(solvator.box_dim)
# CHARMM scripting: crystal build cutoff @boxhalf noper 0
crystal.build(solvator.box_dim/2)

# Turn on image centering - bysegment for protein, by residue for solvent and ions
# CHARMM scripting: image byseg xcen 0 ycen 0 zcen 0 select segid SEGID end
for segid in non_solvent_segids:
    image.setup_segment(0.0, 0.0, 0.0, segid)
# CHARMM scripting: image byres xcen 0 ycen 0 zcen 0 select resname tip3 end
image.setup_residue(0.0, 0.0, 0.0, 'TIP3')
# CHARMM scripting: image byres xcen 0 ycen 0 zcen 0 select resname ion_type end
for ion_type in all_ion_types:
    image.setup_residue(0.0, 0.0, 0.0, ion_type)

# Now specify nonbonded cutoffs for solvated box
cutnb = min(solvator.box_dim/2, 12)
cutim = cutnb
ctofnb = cutnb - 1.0
ctonnb = cutnb - 3.0

# Another nbonds example
# CHARMM scripting: nbonds cutnb @cutnb cutim @cutim ctofnb @ctofnb ctonnb @ctonnb -
#        inbfrq -1 imgfrq -1
non_bonded_script = pycharmm.NonBondedScript(
    cutnb=cutnb, cutim=cutim, ctonnb=ctonnb, ctofnb=ctofnb,
    eps=1.0,
    cdie=True,
    atom=True, vatom=True,
    fswitch=True, vfswitch=True,
    inbfrq=-1, imgfrq=-1
)

# We want to fix the protein and ligands and minimize the solvent to "fit"
# Select everything but solvent and ions
cons_fix_atoms = pycharmm.SelectAtoms()
for segid in non_solvent_segids:
    cons_fix_atoms |= pycharmm.SelectAtoms(seg_id=segid)

# Minimize the solvent positions with periodic boundary conditions using steepest descents
ener_dict = sd_minimize(200, non_bonded_script, cons_fix_selection=cons_fix_atoms)

fetch_coords_from_charmm(model)
pcm_write.coor_card(f'{PDBID}.crd')
pcm_write.psf_card(f'{PDBID}.psf')



"""
Module for reading CHARMM PSF (Protein Structure File) format files.

This module provides functions to parse CHARMM PSF files and return
structured topology data that can be used for comparison or integration
with crimm structures.

Format specification (from CHARMM source io/psfres.F90):
- Extended format (EXT): I10 for integers, A8 for strings
- Standard format: I8 for integers, A4 for strings
- XPLOR format: Uses atom type names instead of parameter file indices
"""

from typing import List, Tuple, Dict, Any, NamedTuple
from dataclasses import dataclass, field


class PSFAtom(NamedTuple):
    """Atom record from PSF file."""
    serial: int
    segid: str
    resid: str
    resname: str
    atomname: str
    atomtype: str
    charge: float
    mass: float
    imove: int = 0


@dataclass
class PSFData:
    """Container for parsed PSF topology data.

    Attributes
    ----------
    title : List[str]
        Title lines from PSF file
    atoms : List[PSFAtom]
        Atom records
    bonds : List[Tuple[int, int]]
        Bond pairs (1-based atom indices)
    angles : List[Tuple[int, int, int]]
        Angle triplets (1-based atom indices)
    dihedrals : List[Tuple[int, int, int, int]]
        Dihedral quadruplets (1-based atom indices)
    impropers : List[Tuple[int, int, int, int]]
        Improper quadruplets (1-based atom indices)
    donors : List[Tuple[int, int]]
        H-bond donor pairs
    acceptors : List[Tuple[int, int]]
        H-bond acceptor pairs
    nonbonded : List[int]
        Non-bonded exclusion indices
    groups : List[Tuple[int, int, int]]
        Atom group definitions (pointer, type, move_flag)
    cmap : List[Tuple[int, ...]]
        CMAP cross-term definitions (8 atom indices each)
    lonepairs : List[Dict[str, Any]]
        Lone pair definitions
    extended : bool
        Whether file was in extended format
    xplor : bool
        Whether file was in XPLOR format
    has_cmap : bool
        Whether file contains CMAP terms
    """
    title: List[str] = field(default_factory=list)
    atoms: List[PSFAtom] = field(default_factory=list)
    bonds: List[Tuple[int, int]] = field(default_factory=list)
    angles: List[Tuple[int, int, int]] = field(default_factory=list)
    dihedrals: List[Tuple[int, int, int, int]] = field(default_factory=list)
    impropers: List[Tuple[int, int, int, int]] = field(default_factory=list)
    donors: List[Tuple[int, int]] = field(default_factory=list)
    acceptors: List[Tuple[int, int]] = field(default_factory=list)
    nonbonded: List[int] = field(default_factory=list)
    groups: List[Tuple[int, int, int]] = field(default_factory=list)
    cmap: List[Tuple[int, ...]] = field(default_factory=list)
    lonepairs: List[Dict[str, Any]] = field(default_factory=list)
    extended: bool = False
    xplor: bool = False
    has_cmap: bool = False


class PSFReader:
    """Read CHARMM PSF files into structured data.

    Attributes
    ----------
    extended : bool
        Whether file uses extended format
    xplor : bool
        Whether file uses XPLOR format
    has_cmap : bool
        Whether file contains CMAP terms
    """

    def __init__(self):
        self.extended = False
        self.xplor = False
        self.has_cmap = False
        self._lines: List[str] = []
        self._line_idx: int = 0

    def read(self, filepath: str) -> PSFData:
        """Parse PSF file and return topology data.

        Parameters
        ----------
        filepath : str
            Path to PSF file

        Returns
        -------
        PSFData
            Parsed topology data
        """
        with open(filepath, 'r', encoding='utf-8') as f:
            self._lines = [line.rstrip() for line in f.readlines()]

        self._line_idx = 0
        data = PSFData()

        # Parse header
        self._parse_header()
        data.extended = self.extended
        data.xplor = self.xplor
        data.has_cmap = self.has_cmap

        # Parse sections
        while self._line_idx < len(self._lines):
            line = self._current_line()

            if not line or line.startswith('!'):
                self._line_idx += 1
                continue

            if '!NTITLE' in line:
                data.title = self._parse_title(line)
            elif '!NATOM' in line:
                data.atoms = self._parse_atoms(line)
            elif '!NBOND' in line:
                data.bonds = self._parse_bonds(line)
            elif '!NTHETA' in line:
                data.angles = self._parse_angles(line)
            elif '!NPHI' in line:
                data.dihedrals = self._parse_dihedrals(line)
            elif '!NIMPHI' in line:
                data.impropers = self._parse_impropers(line)
            elif '!NDON' in line:
                data.donors = self._parse_pairs(line)
            elif '!NACC' in line:
                data.acceptors = self._parse_pairs(line)
            elif '!NNB' in line:
                data.nonbonded = self._parse_nonbonded(line)
            elif '!NGRP' in line:
                data.groups = self._parse_groups(line)
            elif '!NCRTERM' in line:
                data.cmap = self._parse_cmap(line)
            elif '!NUMLP' in line:
                data.lonepairs = self._parse_lonepairs(line)
            else:
                self._line_idx += 1

        return data

    def _current_line(self) -> str:
        """Get current line."""
        if self._line_idx < len(self._lines):
            return self._lines[self._line_idx]
        return ""

    def _next_line(self) -> str:
        """Advance and return next line."""
        self._line_idx += 1
        return self._current_line()

    def _parse_header(self) -> None:
        """Parse PSF header line to detect format flags."""
        line = self._current_line().upper()

        if not line.startswith('PSF'):
            raise ValueError("File does not appear to be a PSF file (missing PSF header)")

        self.extended = 'EXT' in line
        self.xplor = 'XPLOR' in line
        self.has_cmap = 'CMAP' in line

        self._line_idx += 1

    def _parse_count(self, line: str) -> int:
        """Parse count from section header line."""
        # Extract number before !LABEL
        parts = line.split('!')
        count_str = parts[0].strip()
        # Handle multiple numbers (e.g., NGRP NST2)
        nums = count_str.split()
        return int(nums[0]) if nums else 0

    def _parse_title(self, line: str) -> List[str]:
        """Parse NTITLE section."""
        count = self._parse_count(line)
        titles = []
        for _ in range(count):
            self._next_line()
            titles.append(self._current_line().strip())
        self._line_idx += 1
        return titles

    def _parse_atoms(self, line: str) -> List[PSFAtom]:
        """Parse NATOM section."""
        count = self._parse_count(line)
        atoms = []

        for _ in range(count):
            self._next_line()
            atom_line = self._current_line()
            atom = self._parse_atom_line(atom_line)
            atoms.append(atom)

        self._line_idx += 1
        return atoms

    def _parse_atom_line(self, line: str) -> PSFAtom:
        """Parse a single atom line."""
        # Split by whitespace for XPLOR format
        parts = line.split()

        if len(parts) < 9:
            raise ValueError(f"Invalid atom line: {line}")

        serial = int(parts[0])
        segid = parts[1]
        resid = parts[2]
        resname = parts[3]
        atomname = parts[4]
        atomtype = parts[5]
        charge = float(parts[6])
        mass = float(parts[7])
        imove = int(parts[8]) if len(parts) > 8 else 0

        return PSFAtom(
            serial=serial,
            segid=segid,
            resid=resid,
            resname=resname,
            atomname=atomname,
            atomtype=atomtype,
            charge=charge,
            mass=mass,
            imove=imove
        )

    def _read_integers(self, count: int, ints_per_element: int) -> List[int]:
        """Read specified number of integers from following lines."""
        total_ints = count * ints_per_element
        integers = []

        while len(integers) < total_ints:
            self._next_line()
            line = self._current_line()
            if not line or line.startswith('!') or '!' in line:
                break
            # Parse all integers from line
            nums = line.split()
            integers.extend(int(n) for n in nums)

        return integers[:total_ints]

    def _parse_bonds(self, line: str) -> List[Tuple[int, int]]:
        """Parse NBOND section."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 2)

        bonds = []
        for i in range(0, len(integers), 2):
            bonds.append((integers[i], integers[i + 1]))

        self._line_idx += 1
        return bonds

    def _parse_angles(self, line: str) -> List[Tuple[int, int, int]]:
        """Parse NTHETA section."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 3)

        angles = []
        for i in range(0, len(integers), 3):
            angles.append((integers[i], integers[i + 1], integers[i + 2]))

        self._line_idx += 1
        return angles

    def _parse_dihedrals(self, line: str) -> List[Tuple[int, int, int, int]]:
        """Parse NPHI section."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 4)

        dihedrals = []
        for i in range(0, len(integers), 4):
            dihedrals.append((
                integers[i], integers[i + 1],
                integers[i + 2], integers[i + 3]
            ))

        self._line_idx += 1
        return dihedrals

    def _parse_impropers(self, line: str) -> List[Tuple[int, int, int, int]]:
        """Parse NIMPHI section."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 4)

        impropers = []
        for i in range(0, len(integers), 4):
            impropers.append((
                integers[i], integers[i + 1],
                integers[i + 2], integers[i + 3]
            ))

        self._line_idx += 1
        return impropers

    def _parse_pairs(self, line: str) -> List[Tuple[int, int]]:
        """Parse NDON or NACC section (pairs of integers)."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 2)

        pairs = []
        for i in range(0, len(integers), 2):
            pairs.append((integers[i], integers[i + 1]))

        self._line_idx += 1
        return pairs

    def _parse_nonbonded(self, line: str) -> List[int]:
        """Parse NNB section."""
        count = self._parse_count(line)

        # NNB contains exclusion list pointers (one per atom)
        integers = []
        while len(integers) < count:
            self._next_line()
            curr_line = self._current_line()
            if not curr_line or '!' in curr_line:
                break
            nums = curr_line.split()
            integers.extend(int(n) for n in nums)

        self._line_idx += 1
        return integers[:count]

    def _parse_groups(self, line: str) -> List[Tuple[int, int, int]]:
        """Parse NGRP NST2 section."""
        # Line format: NGRP NST2 !NGRP NST2
        parts = line.split('!')
        nums = parts[0].split()
        ngrp = int(nums[0]) if nums else 0

        integers = self._read_integers(ngrp, 3)

        groups = []
        for i in range(0, len(integers), 3):
            groups.append((integers[i], integers[i + 1], integers[i + 2]))

        self._line_idx += 1
        return groups

    def _parse_cmap(self, line: str) -> List[Tuple[int, ...]]:
        """Parse NCRTERM section (cross-map terms)."""
        count = self._parse_count(line)
        integers = self._read_integers(count, 8)

        cmaps = []
        for i in range(0, len(integers), 8):
            cmaps.append(tuple(integers[i:i + 8]))

        self._line_idx += 1
        return cmaps

    def _parse_lonepairs(self, line: str) -> List[Dict[str, Any]]:
        """Parse NUMLP NUMLPH section."""
        parts = line.split('!')
        nums = parts[0].split()
        nlp = int(nums[0]) if nums else 0

        lonepairs = []
        for _ in range(nlp):
            self._next_line()
            lp_line = self._current_line()
            parts = lp_line.split()
            if len(parts) >= 6:
                lonepairs.append({
                    'host': int(parts[0]),
                    'lp': int(parts[1]),
                    'type': parts[2],
                    'distance': float(parts[3]),
                    'angle': float(parts[4]),
                    'dihedral': float(parts[5])
                })

        self._line_idx += 1
        return lonepairs


# Convenience function

def read_psf(filepath: str) -> PSFData:
    """Read CHARMM PSF file and return parsed data.

    Parameters
    ----------
    filepath : str
        Path to PSF file

    Returns
    -------
    PSFData
        Parsed topology data
    """
    reader = PSFReader()
    return reader.read(filepath)


def compare_psf(psf1: PSFData, psf2: PSFData, verbose: bool = False) -> Dict[str, Any]:
    """Compare two PSF data structures.

    Parameters
    ----------
    psf1 : PSFData
        First PSF data
    psf2 : PSFData
        Second PSF data
    verbose : bool, default False
        Print detailed comparison results

    Returns
    -------
    Dict[str, Any]
        Comparison results with keys: 'equal', 'differences'
    """
    differences = []

    # Compare atom counts
    if len(psf1.atoms) != len(psf2.atoms):
        differences.append(f"Atom count: {len(psf1.atoms)} vs {len(psf2.atoms)}")

    # Compare atoms (if same count)
    if len(psf1.atoms) == len(psf2.atoms):
        for i, (a1, a2) in enumerate(zip(psf1.atoms, psf2.atoms)):
            if a1.atomname != a2.atomname:
                differences.append(f"Atom {i+1} name: {a1.atomname} vs {a2.atomname}")
            if a1.atomtype != a2.atomtype:
                differences.append(f"Atom {i+1} type: {a1.atomtype} vs {a2.atomtype}")
            if abs(a1.charge - a2.charge) > 1e-4:
                differences.append(f"Atom {i+1} charge: {a1.charge} vs {a2.charge}")
            if abs(a1.mass - a2.mass) > 1e-4:
                differences.append(f"Atom {i+1} mass: {a1.mass} vs {a2.mass}")

    # Compare topology counts
    counts = [
        ('bonds', len(psf1.bonds), len(psf2.bonds)),
        ('angles', len(psf1.angles), len(psf2.angles)),
        ('dihedrals', len(psf1.dihedrals), len(psf2.dihedrals)),
        ('impropers', len(psf1.impropers), len(psf2.impropers)),
        ('cmap', len(psf1.cmap), len(psf2.cmap)),
    ]

    for name, c1, c2 in counts:
        if c1 != c2:
            differences.append(f"{name.capitalize()} count: {c1} vs {c2}")

    # Compare bond sets (order-independent)
    bonds1 = set(tuple(sorted(b)) for b in psf1.bonds)
    bonds2 = set(tuple(sorted(b)) for b in psf2.bonds)
    if bonds1 != bonds2:
        missing1 = bonds2 - bonds1
        missing2 = bonds1 - bonds2
        if missing1:
            differences.append(f"Bonds in psf2 but not psf1: {len(missing1)}")
        if missing2:
            differences.append(f"Bonds in psf1 but not psf2: {len(missing2)}")

    result = {
        'equal': len(differences) == 0,
        'differences': differences
    }

    if verbose:
        if result['equal']:
            print("PSF files are equivalent")
        else:
            print("PSF files differ:")
            for diff in differences[:20]:  # Limit output
                print(f"  - {diff}")
            if len(differences) > 20:
                print(f"  ... and {len(differences) - 20} more differences")

    return result



import warnings
from copy import copy
import requests
import pandas as pd
from Bio.PDB.Selection import unfold_entities
# Nucleoside phosphates and phosphonates
from crimm.Data.components_dict import NUCLEOSIDE_PHOS, PDB_CHARMM_ION_NAMES, CHARMM_PDB_ION_NAMES 
from crimm.Utils.query_db import query_drugbank_info
from crimm.Utils.StructureUtils import index_to_letters, letters_to_index
from .Model import Model
from .Chain import (
    PolymerChain, Heterogens, Solvent, CoSolvent, Ion,
    Glycosylation, NucleosidePhosphate, Chain, Ligand, Oligosaccharide, Macrolide
)
from .Residue import Residue

class OrganizedModel(Model):
    """The OrganizedModel class represents a model in a structure with a specific 
    organization of chains. It is derived from the Model class and is used in the 
    Structure class to organize chains into categories. Require network access to
    fetch binding affinity and drugbank information. Alt locs will be removed and only 
    the selected alt loc atom in disordered residues will be kept in the model.
    The following chain types are supported:
    - protein: Polypeptide(L)
    - rna: Polyribonucleotide
    - dna: Polydeoxyribonucleotide
    - other_polymer: PolymerChain
    - solvent: Solvent
    - unknown_type: UnknownType
    - phos_ligand: NucleosidePhosphate
    - ligand: Ligand
    - co_solvent: CoSolvent
    - ion: Ion
    - glycosylation: Glycosylation
    - heterogens: Heterogens
    

    """
    chain_types = {
        'Polypeptide(L)': PolymerChain,
        'Polyribonucleotide': PolymerChain,
        'Polydeoxyribonucleotide': PolymerChain,
        'Oligosaccharide': Oligosaccharide,
        'Macrolide': Macrolide,
        'Solvent': Solvent,
        'Heterogens': Heterogens,
        'Ligand': Ligand,
        'CoSolvent': CoSolvent,
        'Ion': Ion,
        'Glycosylation': Glycosylation,
        'NucleosidePhosphate': NucleosidePhosphate,
        'UnknownType': Chain,
    }
    organized_chains = {
        'protein': 'Polypeptide(L)',
        'rna': 'Polyribonucleotide',
        'dna': 'Polydeoxyribonucleotide',
        'oligosaccharide': 'Oligosaccharide',
        'macrolide': 'Macrolide',
        'other_polymer': 'PolymerChain',
        'solvent': 'Solvent',
        'unknown_type': 'UnknownType',
        'phos_ligand': 'NucleosidePhosphate',
        'ligand': 'Ligand',
        'co_solvent': 'CoSolvent',
        'ion': 'Ion',
        'glycosylation': 'Glycosylation',
    }
    def __init__(
            self, entity, rename_charmm_ions=True, rename_solvent_oxygen=True,
            identify_ligands=False
        ):
        """Initialize the OrganizedModel object.
        Args:
            entity (Entity): The entity to be organized.
            rename_charmm_ions (bool): Whether to rename ions in the structure to 
            CHARMM ion name defined in water_ions.str
            rename_solvent_oxygen (bool): Whether to rename solvent oxygen to CHARMM 
            name "OH2" in the crystallographic water. Doing so will allow crimm to 
            generate topology definitions on the water
        """
        if entity.level == 'M':
            model = entity
            pdbx_description = entity.pdbx_description
        elif entity.level == 'S':
            model = entity.models[0]
            pdbx_description = entity.models[0].pdbx_description
        else:
            raise ValueError(
                "Only Structure level (S) or Model level (M) entity is accepted for OrganizedModel!"
                f"{entity}  has level \"{entity.level}\"."
            )
        super().__init__(model.id)
        self.pdbx_description = pdbx_description
        self.connect_dict = copy(model.connect_dict)
        ## keep original model's connect atoms for looking up covalent bonds
        ## for glycosylation
        self._ref_connect_atoms = model.connect_atoms
        self.identify_ligands = identify_ligands
        self.rcsb_web_data = None
        # Binding Affinity Information
        self.binding_info = None
        # Biologically Interesting Molecules Information
        self.bio_mol_info = None
        # Names (3-letter code) for ligand with binding affinity information
        self.lig_names = set()
        # Names (descriptor string) for ligand that exists in 
        # Biologically Interesting Molecules databases
        self.bio_mol_names = set()
        self.topology_loader = None
        
        if model.pdb_id is not None:
            self.pdb_id = model.pdb_id
            self._get_rcsb_web_data()
            self.binding_info = self._get_keyword_info('rcsb_binding_affinity')
            self.bio_mol_info = self._get_keyword_info('pdbx_molecule_features')
        elif model.parent.id is not None:
            self.pdb_id = model.parent.id
            self._get_rcsb_web_data()
            self.binding_info = self._get_keyword_info('rcsb_binding_affinity')
            self.bio_mol_info = self._get_keyword_info('pdbx_molecule_features')
        else:
            warnings.warn(
                "PDB ID not set for this model! "
                "Binding affinity information will not be fetched. Possible errors "
                "in ligand classification."
            )

        if self.binding_info is not None:
            self.lig_names = set(self.binding_info.comp_id)
        if self.bio_mol_info is not None:
            self.bio_mol_names = set(self.bio_mol_info.name)
        self.organize(model, identify_ligands=self.identify_ligands)
        if rename_solvent_oxygen:
            self.rename_solvent_oxygen()
        if rename_charmm_ions:
            self.rename_charmm_ions()
        
        
    
    def filter(self, name):
        """Return the type of the chain with the given name."""
        name = name.rstrip('s').lower()
        if name in self.organized_chains:
            chains = []
            for chain in self:
                if chain.chain_type == self.organized_chains[name]:
                    chains.append(chain)
            return chains
        raise KeyError(f"Invalid chain type {name}! Available types are {self.organized_chains.keys()}.")

    @property
    def protein(self):
        return self.filter('protein')
    @property
    def RNA(self):
        return self.filter('rna')
    @property
    def DNA(self):
        return self.filter('dna')
    @property
    def oligosaccharide(self):
        return self.filter('oligosaccharide')
    @property
    def macrolide(self):
        return self.filter('macrolide')
    @property
    def other_polymer(self):
        return self.filter('other_polymer')
    @property
    def solvent(self):
        return self.filter('solvent')
    @property
    def unknown_type(self):
        return self.filter('unknown_type')
    @property
    def phos_ligand(self):
        return self.filter('phos_ligand')
    @property
    def ligand(self):
        return self.filter('ligand')
    @property
    def co_solvent(self):
        return self.filter('co_solvent')
    @property
    def ion(self):
        return self.filter('ion')
    @property
    def glycosylation(self):
        return self.filter('glycosylation')
    
    @property
    def non_solvent(self):
        """Return all chains that are not solvent chains."""
        non_solvent = []
        for chain in self:
            if chain.chain_type != 'Solvent':
                non_solvent.append(chain)
        return non_solvent
    
    def _get_rcsb_web_data(self):
        query_url = f'https://data.rcsb.org/rest/v1/core/entry/{self.pdb_id}'
        response = requests.get(query_url, timeout=500)
        if (code := response.status_code) != 200:
            warnings.warn(
                f"GET request on RCSB for \"{self.pdb_id}\" for binding affinity data "
                f"did not return valid result! \n[Status Code] {code}"
            )
            return
        self.rcsb_web_data = response.json()

    def _get_keyword_info(self, keyword):
        if self.rcsb_web_data is None or keyword not in self.rcsb_web_data:
            return
        return pd.DataFrame(self.rcsb_web_data[keyword])

    def is_glycosylation(self, residue):
        """Check if the heterogen is part of glycosylation (covalently bonded sugar)"""
        if self._ref_connect_atoms is None:
            return False
        if 'covale' not in self._ref_connect_atoms:
            return False
        for atom_pair in self._ref_connect_atoms['covale']:
            if residue in unfold_entities(atom_pair, 'R'):
                return True
        return False

    def is_ligand(self, residue):
        """Check if the residue is a ligand."""
        if len(self.lig_names) > 0:
            return residue.resname in self.lig_names
        elif len(self.bio_mol_names) > 0:
            return residue.pdbx_description in self.bio_mol_names
        else:
            return query_drugbank_info(residue.resname) is not None
    
    def create_hetero_chain(
            self, chain_id, hetero_res_list, chain_type='Heterogens', 
            reset_resid=True
        ):
        """Create a new Heterogens chain from a list of heterogens."""
        all_description = []
        hetero_chain = self.chain_types[chain_type](chain_id)
        resnames = set()
        for i, res in enumerate(hetero_res_list, start=1):
            description = None
            if hasattr(res, 'pdbx_description'):
                description = res.pdbx_description
            elif res.parent is not None:
                description  = res.parent.pdbx_description
            if description not in all_description and description is not None:
                all_description.append(description)
            res.detach_parent()
            if reset_resid:
                het_flag, resseq, icode = res.id
                res.id = (het_flag, i, icode)
            hetero_chain.add(res)
            resnames.add(res.resname)

        if chain_type == 'Solvent':
            hetero_chain.pdbx_description = 'water'
        elif len(all_description) > 0:
            hetero_chain.pdbx_description = ', '.join(all_description)
        hetero_chain.resnames = ', '.join(resnames)
        return hetero_chain

    def determine_heterogen_type(self, heterogens, identify_ligands=False):
        """Determine the type of heterogens in the model."""
        heterogen_type_dict = {
            'Ion': [],
            'Glycosylation': [],
            'Ligand': [],
            'NucleosidePhosphate': [],
            'CoSolvent': []   
        }
        
        for res in heterogens:
            if len(res.resname) <= 2 or res.resname in CHARMM_PDB_ION_NAMES:
                heterogen_type_dict['Ion'].append(res)
            elif self.is_glycosylation(res):
                heterogen_type_dict['Glycosylation'].append(res)
            elif not identify_ligands:
                # If the ligand is not guessed, all heterogens except ions and
                # identified glycosylation will be considered ligands
                heterogen_type_dict['Ligand'].append(res)
            elif res.resname in NUCLEOSIDE_PHOS:
                heterogen_type_dict['NucleosidePhosphate'].append(res)
            elif self.is_ligand(res):
                heterogen_type_dict['Ligand'].append(res)
            else:
                heterogen_type_dict['CoSolvent'].append(res)
        return heterogen_type_dict
    
    def update(self):
        self.organize(self, identify_ligands=self.identify_ligands)

    def organize(self, model: Model, identify_ligands=False):
        """Organize the chains in the model into categories."""
        undecided_heterogens = []
        all_chains = []
        solvent_entry = {'Solvent': []}
        chain_id_map = {}
        for chain in model.chains:
            if chain.chain_type == 'Solvent':
                solvent_entry['Solvent'].extend(chain.residues)
            elif chain.chain_type == 'Heterogens':
                undecided_heterogens.extend(chain.residues)
            else:
                chain_id_map[chain.id] = chain.id
                all_chains.append(chain)

        heterogen_type_dict = self.determine_heterogen_type(
            undecided_heterogens, identify_ligands=identify_ligands
        )
        heterogen_type_dict.update(solvent_entry)
        named_hetero_chains = {}
        for chain_type, hetero_res_list in heterogen_type_dict.items():
            named_hetero_chains[chain_type] = {}
            cur_chain_type = named_hetero_chains[chain_type]
            for res in hetero_res_list:
                temp_id = f'_{res.resname}'
                chain_id_map[res.parent.id] = temp_id
                if temp_id in cur_chain_type:
                    cur_chain_type[temp_id].append(res)
                else:
                    cur_chain_type[temp_id] = [res]
        for chain_type, resname_dict in named_hetero_chains.items():
            if len(resname_dict) == 0:
                continue
            for temp_id, hetero_res_list in resname_dict.items():
                hetero_chain = self.create_hetero_chain(
                    temp_id, hetero_res_list, chain_type
                )
                all_chains.append(hetero_chain)

        all_chains.sort(key=lambda x: x.id)
        temp_id_map = {}
        for i, chain in enumerate(all_chains):
            chain.detach_parent()
            temp_id = chain.id
            new_id = index_to_letters(i)
            temp_id_map[temp_id] = new_id
            chain.id = new_id
            self.add(chain)
        chain_id_map = {k: temp_id_map[v] for k, v in chain_id_map.items()}
        # update the connected atoms in the model level bonds (e.g. disulfide bonds)
        self._update_connect_dict_chain_id(chain_id_map)
        self.set_connect(self.connect_dict)
            
    def _update_connect_dict_chain_id(self, id_map: dict):
        """Update the chain ID in the connect_dict."""
        for entries in self.connect_dict.values():
            for atom_pair_dicts in entries:
                for atom_info in atom_pair_dicts:
                    if atom_info['chain'] in id_map:
                        atom_info['chain'] = id_map[atom_info['chain']]
                    else:
                        print(f"Chain {atom_info['chain']} not found in the model!")

    def rename_solvent_oxygen(self):
        """Rename oxygen atom in solvent to OH2."""
        for chain in self.solvent:
            for res in chain:
                if 'O' in res:
                    atom = res['O']
                    atom.rename('OH2')

    def replace_ion(self, res, new_ion_name):
        """Replace the ion residue by changing its name."""
        new_ion_name = new_ion_name.upper()
        if res.resname == new_ion_name:
            return
        het_flag, resseq, icode = res.id
        het_flag = f'H_{new_ion_name}'
        res.id = (het_flag, resseq, icode)
        res.resname = new_ion_name
        for atom in res:
            # Rename the atom. Atom name should be the same as resname here
            atom.rename(new_ion_name)

    def replace_ion_chain(self, ion_chain, new_ion_name):
        """Replace the ion chain by changing its name."""
        if ion_chain.chain_type != 'Ion':
            raise ValueError(
                f"Chain {ion_chain.id} is not an Ion chain!"
            )
        new_ion_name = new_ion_name.upper()
        for res in ion_chain:
            self.replace_ion(res, new_ion_name)
        ion_chain.resnames = new_ion_name
        ion_chain.pdbx_description = new_ion_name

    def rename_charmm_ions(self):
        """Rename the ions to CHARMM naming convention."""
        for chain in self.ion:
            for res in chain:
                charmm_name = PDB_CHARMM_ION_NAMES.get(
                        res.resname.upper(), res.resname
                    )
                if res.resname == 'ZN':
                    res.resname = 'ZN2'
                    continue
                elif res.resname == charmm_name:
                    continue
                self.replace_ion(res, charmm_name)

    def __repr__(self):
        repr_str = f"<OrganizedModel model={self.pdb_id} "
        for key, val in self.organized_chains.items():
            if len(self.filter(key)) > 0:
                repr_str += f"{val}={len(self.filter(key))} "
        repr_str += ">"
        return repr_str

    def _ipython_display_(self):
        """Return the nglview interactive visualization window"""
        if len(self) != 0:
            from crimm.Visualization import show_nglview_multiple
            from IPython.display import display
            display(show_nglview_multiple(self.child_list))
        print(self.expanded_view())
        
    def correct_chain_type(self, chain_id, chain_type):
        """Reassign the chain type of a chain in the model."""
        if (general_name:=chain_type.rstrip('s').lower()) in self.organized_chains:
            ## Use the actual chain class name if user provides the organized_chains name
            chain_type = self.organized_chains[general_name]
        if chain_type not in self.chain_types:
            raise ValueError(f"Invalid chain type {chain_type}!")
        old_chain = self[chain_id]
        if chain_type == old_chain.chain_type:
            return
        self.detach_child(chain_id)
        new_chain = self.create_hetero_chain(
            old_chain.id, old_chain.residues, chain_type
        )
        self.add(new_chain)

    def combine_chains_by_id(self, chain_ids, new_id=None):
        """Combine the chains in the model into a single chain."""
        if len(chain_ids) == 0:
            return
        chains = [self[chain_id] for chain_id in chain_ids]
        self.combine_chains(chains, new_id)


    def combine_chains(self, chains, new_id=None):
        """Combine a list of chains into a single chain."""
        if len(chains) == 0:
            return
        chain_types = set(c.chain_type for c in chains)
        if len(chain_types) > 1:
            raise ValueError(
                f"Cannot combine chains with different chain types! {chain_types}"
                "found in the chain list."
            )
        chain_type = chain_types.pop()
        if new_id is None:
            new_id = self[0].id
        new_chain = self.chain_types[chain_type](new_id)
        all_res = [res for chain in chains for res in chain]
        for i, res in enumerate(all_res, start=1):
            het_flag, resseq, icode = res.id
            res.id = (het_flag, i, icode)
            new_chain.add(res)

        for chain in chains:
            self.detach_child(chain.id)
        self.add(new_chain)

    def _add_ligand_residue(self, res, chain_id=None, description=None):
        """Add a ligand residue to the model."""
        self.sort_chains(reset_id=False)
        if description is not None:
            res.pdbx_description = description
        if chain_id in self:
            chain = self[chain_id]
            chain.add(res)
            all_descriptions = set(res.pdbx_description for res in chain if res.pdbx_description is not None)
            if len(all_descriptions) > 0:
                chain.pdbx_description = ', '.join(all_descriptions)
            else:
                chain.pdbx_description = None
            return

        if chain_id is None:
            chain_id = index_to_letters(letters_to_index(self.chains[-1].id)+1)
            assert chain_id not in self
        
        chain = self.chain_types['Ligand'](chain_id)
        chain.pdbx_description = res.pdbx_description
        chain.resnames = [res.resname]
        self.add(chain)
        chain.add(res)
        
    def _add_ligand_chain(self, chain, description=None):
        """Add a ligand chain to the model."""
        self.sort_chains(reset_id=False)
        if chain.id in self:
            new_id = index_to_letters(letters_to_index(self.chains[-1].id)+1)
            warnings.warn(
                f"Chain {chain.id} already exists in the model! "
                f"Ligand chain id will be renamed to {new_id}."
            )
            chain.id = new_id
        self.add(chain)
        if description is not None:
            chain.pdbx_description = description

    def add_ligand(self, ligand, chain_id=None, description=None):
        """Add a ligand to the model."""
        if isinstance(ligand, Residue):
            self._add_ligand_residue(ligand, chain_id, description)
        elif isinstance(ligand, Chain):
            ligand.id = chain_id
            self._add_ligand_chain(ligand, description)
        else:
            raise ValueError(f"Invalid ligand type {type(ligand)}!")


import warnings
from collections import OrderedDict
import numpy as np
import numpy.linalg as LA
from scipy.spatial.transform import Rotation as R

def sep_by_priorities(atom_list):
    # Create a 2-tier priotity list to separate
    # heavy atoms and hydrogen atom
    missing_atoms, missing_hydrogens = [],[]
    for atom in atom_list:
        if atom.startswith('H'):
            missing_hydrogens.append(atom)
        else:
            missing_atoms.append(atom)
    return missing_atoms, missing_hydrogens
    
def recur_find_build_seq(
    running_dict: OrderedDict, missing_atoms, build_seq, exclude_list
):

    if len(missing_atoms) == 0:
        return build_seq
 
    atom_name, ic_list = running_dict.popitem(last=False)
    for ic_keys in ic_list:
        cur_ic_set = set(ic_keys)
        if (len(cur_ic_set.intersection(missing_atoms)) > 1) or (
            cur_ic_set.intersection(exclude_list)
        ):
            continue

        missing_atoms.remove(atom_name)
        build_seq.append((atom_name, ic_keys))
        return recur_find_build_seq(
            running_dict, missing_atoms, build_seq, exclude_list
        )

    running_dict[atom_name] = ic_list
    return recur_find_build_seq(
        running_dict, missing_atoms, build_seq, exclude_list
    )

def get_coord_from_dihedral_ic(i_coord, j_coord, k_coord, phi, t_jkl, r_kl):
    origin = j_coord
    a1 = i_coord - origin
    a2 = k_coord - origin
    r1 = R.from_rotvec(phi * a2/LA.norm(a2), degrees=True)
    n1 = np.cross(a1, a2)
    n2 = r1.apply(n1)
    m = np.cross(a2, n2)
    r2 = R.from_rotvec((t_jkl-90) * n2/LA.norm(n2), degrees=True)
    a3 = r2.apply(m)
    a3 = a3/LA.norm(a3)*r_kl

    l_coord = origin+a2+a3
    return l_coord
    
def get_coord_from_improper_ic(i_coord, j_coord, k_coord, phi, t_jkl, r_kl):
    origin = k_coord
    a1 = origin - i_coord
    a2 = origin - j_coord
    r1 = R.from_rotvec(phi * a2/LA.norm(a2), degrees=True)
    n1 = np.cross(a2, a1)
    n2 = r1.apply(n1)
    m = np.cross(n2, a2)
    r2 = R.from_rotvec((t_jkl+90) * n2/LA.norm(n2), degrees=True)
    a3 = r2.apply(m)
    a3 = a3/LA.norm(a3)*r_kl
    
    l_coord = origin+a3
    return l_coord
    
def find_build_seq(topo_def, missing_atoms, missing_hydrogens):
    missing_atoms = list(missing_atoms)
    missing_hydrogens = list(missing_hydrogens)
    all_missing = missing_atoms + missing_hydrogens
    lookup_dict = topo_def.atom_lookup_dict
    running_dict = OrderedDict({k: lookup_dict[k] for k in all_missing})

    heavy_build_seq = recur_find_build_seq(
        running_dict,
        missing_atoms.copy(), # Prevent element removal on the original list
        build_seq = [], 
        exclude_list = missing_hydrogens
    )

    hydrogen_build_seq = recur_find_build_seq(
        running_dict,
        missing_hydrogens.copy(),
        build_seq = [],
        exclude_list = []
    )

    return heavy_build_seq, hydrogen_build_seq

def find_coords_by_ic(build_sequence, ic_dicts, coord_dict):
    computed_coords = []
    for atom_name, ic_key in build_sequence:
        i, j, k, l = ic_key
        ic_param_dict = ic_dicts[ic_key]
        is_improper = ic_param_dict['improper']
        phi = ic_param_dict['Phi']
        if atom_name == i:
            # the atom is i
            if is_improper:
                # i, j, *k, l => l, *k, j, i = a1, a2, a3, cur_atom
                a1, a2, a3 = coord_dict[j], coord_dict[l], coord_dict[k]
                bond_len = ic_param_dict['R(I-K)']
                bond_angle = ic_param_dict['T(I-K-J)']
                coord = get_coord_from_improper_ic(
                    a1, a2, a3, phi, bond_angle, bond_len
                )
            else:
                # i, j, k, l => l, k, j, i = a1, a2, a3, cur_atom
                a1, a2, a3 = coord_dict[l], coord_dict[k], coord_dict[j]
                bond_len = ic_param_dict['R(I-J)']
                bond_angle = ic_param_dict['T(I-J-K)']
                coord = get_coord_from_dihedral_ic(
                    a1, a2, a3, phi, bond_angle, bond_len
                )
        else:
            # the atom is l
            a1, a2, a3 = coord_dict[i], coord_dict[j], coord_dict[k]
            bond_len = ic_param_dict['R(K-L)']
            bond_angle = ic_param_dict['T(J-K-L)']
            if is_improper:
                coord = get_coord_from_improper_ic(
                    a1, a2, a3, phi, bond_angle, bond_len
                )
            else:
                coord = get_coord_from_dihedral_ic(
                    a1, a2, a3, phi, bond_angle, bond_len
                )
        computed_coords.append(atom_name)
        coord_dict[atom_name] = coord
        
    return computed_coords
    
def ab_initio_ic_build(topo_def):
    ic_dicts = topo_def.ic
    lookup_dict = topo_def.atom_lookup_dict
    all_atoms = [k for k in lookup_dict.keys() if k not in ('-C','CA','N')]
    running_dict = OrderedDict({k: lookup_dict[k] for k in all_atoms})

    build_seq = recur_find_build_seq(
        running_dict,
        all_atoms,
        build_seq = [],
        exclude_list = []
    )

    base_ic1 = ic_dicts[('-C', 'N', 'CA', 'C')]
    base_ic2 = ic_dicts[('N', 'CA', 'C', '+N')]
    r_CN = base_ic1['R(I-J)']
    t_CNCA = np.deg2rad(base_ic1['T(I-J-K)'])
    r_NCA = base_ic2['R(I-J)']

    x_N = np.cos(t_CNCA)*r_NCA
    y_N = np.sin(t_CNCA)*r_NCA

    coord_dict = {
        '-C': np.array([r_CN,0,0], dtype=float),
        'N': np.array([0, 0, 0], dtype=float),
        'CA':np.array([x_N, y_N, 0])
    }
    find_coords_by_ic(build_seq, ic_dicts, coord_dict)

    return coord_dict

class ResidueFixer:
    def __init__(self):
        self._res = None
        self.topo_def = None
        self.heavy_build_sequence = None
        self.hydrogen_build_sequence = None
        self.coord_dict = None
        self.res_type = None
        self.nei_atom_names = None

    @property
    def residue(self):
        """The residue to be fixed."""
        return self._res
    
    @residue.setter
    def residue(self, residue):
        self.load_residue(residue)

    def _create_empty_neighbor_atom_dict_from_ic(self):
        """Get the names of the neighboring atoms for the loaded residue."""
        nei_atoms = {}
        for ic_key in self.topo_def.ic:
            for atom in ic_key:
                if atom.startswith('+') or atom.startswith('-'):
                    nei_atoms[atom] = None
        return nei_atoms
    
    @staticmethod
    def _get_neighbor_atom_coord(atom_names, prev_nei, next_nei):
        """Get the coordinates of the neighboring atoms for list of atom names."""
        coord_dict = {}
        for atom_name in atom_names:
            if atom_name.startswith('+'):
                coord_dict[atom_name] = next_nei[atom_name[1:]].coord
            elif atom_name.startswith('-'):
                coord_dict[atom_name] = prev_nei[atom_name[1:]].coord
            else:
                raise ValueError(
                    'Atom name should start with "+" or "-".'
                )
        return coord_dict
    
    def _get_neighbor_backbone_coords(self):
        """Get the coordinates of the neighboring backbone atoms for the loaded 
        residue."""
        residue = self.residue
        chain = residue.parent
        
        if chain is None:
            warnings.warn(
                'Residue does not belong to a polymer chain, and no neighboring '
                'residues exist. The temperary backbone atoms of the neigbor '
                'residues will be calculated as place holders.'
            )
            return

        resseq = residue.id[1]
        list_idx = chain.child_list.index(residue)
        if (
            (list_idx == 0) or
            (list_idx == len(chain.child_list) - 1)
        ):
            if residue.topo_definition.patch_with is None:
                warnings.warn(
                    'Missing atoms on terminal residues will be built without patching! '
                    'Terminal patching is recommended before building missing atoms!'
                )
            return
        
        prev_res = chain.child_list[list_idx-1]
        next_res = chain.child_list[list_idx+1]
        rev_resseq, next_resseq = prev_res.id[1], next_res.id[1]
        if rev_resseq + 1 != resseq or next_resseq - 1 != resseq:
            warnings.warn(
                'Immediate neighbor residues missing! The temperary backbone atoms '
                'of the neigbor residues will be calculated as place holders.'
            )
            return
        
        neighbor_coord_dict = self._get_neighbor_atom_coord(
            self.nei_atom_names, prev_res, next_res
        )
        return neighbor_coord_dict
    
    def load_residue(self, residue, topo_definition = None):
        """Load the residue to be fixed."""
        if topo_definition is not None:
            self.topo_def = topo_definition
        elif residue.topo_definition is not None:
            self.topo_def = residue.topo_definition
        else:
            raise ValueError(
                'No ResidueDefinition provided nor loaded in the residue! '
                'Residue Topology Definition must be loaded in the residue using '
                'load_topo_definition() or provided as a parameter in the '
                'load_residue() method.'
            )
        self._res = residue
        if residue.resname in self.topo_def.aa_3to1:
            self.res_type = 'aa'
            self.nei_atom_names = ('-C', '+N', '+CA')
        elif residue.resname in ('A', 'G', 'C', 'U', 'T'):
            self.res_type = 'nuc'
            self.nei_atom_names = ("-O3'", "+P", "+O5'")
        else:
            raise ValueError(
                f'Residue {residue} has no neighbor backbone atoms to '
                'build! Due to unkown residue type'
                )
        self.coord_dict = {atom.name: atom.coord for atom in self._res}
        neighbor_coord_dict = self._get_neighbor_backbone_coords()
        if neighbor_coord_dict is None:
            # we only need the name of the missing atoms from neighbors, 
            # not to actually build them
            nei_atom_dict = self._create_empty_neighbor_atom_dict_from_ic()
            residue.missing_atoms.update(nei_atom_dict)
        else:
            self.coord_dict.update(neighbor_coord_dict)
        self.heavy_build_sequence, self.hydrogen_build_sequence = \
            find_build_seq(
            self.topo_def, self.missing_atoms, self.missing_hydrogens
        )

    @property
    def missing_atoms(self):
        """Get the current missing heavy atoms"""
        return self._res.missing_atoms

    @property
    def missing_hydrogens(self):
        """Get the current missing hydrogen atoms"""
        return self._res.missing_hydrogens

    def remove_hydrogens(self):
        """Remove all hydrogens on the residue"""
        hydrogens = []
        for atom in self._res:
            if atom.element == 'H':
                hydrogens.append(atom)
        for hydrogen in hydrogens:
            self._res.detach_child(hydrogen.id)
            if hydrogen in self._res.undefined_atoms:
                self._res.undefined_atoms.remove(hydrogen)
        # update missing hydrogens from the residue topology definition
        remove_h_names = []
        for missing_h in self._res.missing_hydrogens:
            if missing_h not in self._res.topo_definition:
                remove_h_names.append(missing_h)
        for h_name in remove_h_names:
            self._res.missing_hydrogens.pop(h_name)
        for atom_def in self._res.topo_definition:
            if atom_def.element == 'H' and atom_def.name not in self._res.missing_hydrogens:
                self._res.missing_hydrogens[atom_def.name] = atom_def.create_new_atom()
        self.heavy_build_sequence, self.hydrogen_build_sequence = \
            find_build_seq(
            self.topo_def, self.missing_atoms, self.missing_hydrogens
        )


    def _build_atoms(self, build_sequence, missing_atoms:dict):
        computed_atom_names = find_coords_by_ic(
            build_sequence, self.topo_def.ic, self.coord_dict
        )
        built_atoms = []
        for atom_name in computed_atom_names:
            built_atom = missing_atoms.pop(atom_name)
            if atom_name.startswith('+') or atom_name.startswith('-'):
                # neighboring residue's backbone atoms will not be built
                continue
            coord = self.coord_dict[atom_name]
            built_atom.coord = coord
            self._res.add(built_atom)
            built_atoms.append(built_atom)
        return built_atoms
    
    def build_missing_atoms(self):
        """Build missing atoms based on residue topology definition on 
        internal coordinates. If neigbor residue has missing backbone atom that 
        the ic table depends on (namely +N, +CA, and -C), place holders for these 
        atoms will be built first."""
        if len(self.missing_atoms) == 0:
            return
        return self._build_atoms(self.heavy_build_sequence, self.missing_atoms)

    def build_hydrogens(self):
        """Build hydrogens atoms based on residue topology definition on 
        internal coordinates. Any missing heavy atoms will be built before building
        the hydrogens. If neigbor residue has missing backbone atom that the ic table
        depends on (namely +N, +CA, and -C), place holders for these atoms will be 
        built first."""

        if len(self.missing_hydrogens) == 0:
            return
        built_atoms = []
        if len(self.missing_atoms) != 0:
            missing_atom_names = tuple(self.missing_atoms.keys())
            if (
                missing_atom_names != ('-C', '+N', '+CA') or 
                missing_atom_names != ("-O3'", "+P", "+O5'")
            ):
                warnings.warn(
                    f'{len(self.missing_atoms)} Missing heavy atoms are built '
                    f'before building hydrogens: {missing_atom_names}'
                )
            built_atoms.extend(self.build_missing_atoms())
        built_atoms.extend(
            self._build_atoms(
                self.hydrogen_build_sequence, self.missing_hydrogens
            )
        )
        return built_atoms

    def remove_undefined_atoms(self):
        """Remove undefined atoms from the residue"""
        
        detach_ids = [atom.id for atom in self._res.undefined_atoms]
        for atom in self._res.undefined_atoms:
            for atom_group in self._res.atom_groups:
                if atom in atom_group:
                    atom_group.remove(atom)
        for atom_id in detach_ids:
            self._res.detach_child(atom_id)
        self._res.undefined_atoms = []

    def rebuild_hydrogens(self):
        """Rebuild hydrogens on the residue"""
        self.remove_hydrogens()
        built_hydrogens = self.build_hydrogens()
        return built_hydrogens

def build_missing_atoms_for_chain(chain):
    """Build missing residue for a PolymerChain where topology definitions 
    have been loaded for each residue. Missing atom will be built based on the IC
    definitions if the residue is not completely missing (e.g. missing loop on a
    chain).
    
    Args:
        chain: PolymerChain with topology definitions loaded for each residue.
    """
    built_atoms = {}
    res_builder = ResidueFixer()
    chain.sort_residues()
    for res in chain:
        if res.topo_definition is None:
            warnings.warn(f'No topology definition on {res}! Skipped')
            continue
        if res.missing_atoms:
            res_builder.load_residue(res)
            built_atoms[(res.id[1], res.resname)] = res_builder.build_missing_atoms()
    return built_atoms

def build_hydrogens_for_chain(chain, rebuild=False):
    """Build missing hydrogens for a PolymerChain where topology definitions 
    have been loaded for each residue. Missing atom will be built based on the IC
    definitions if the residue is not completely missing (e.g. missing loop on a
    chain).
    
    Args:
        chain: PolymerChain with topology definitions loaded for each residue.
        rebuild: If True, remove all hydrogens and rebuild them. If False, only
            build missing hydrogens.
    """
    built_atoms = {}
    res_builder = ResidueFixer()
    chain.sort_residues()
    for res in chain:
        if res.topo_definition is None:
            warnings.warn(f'No topology definition on {res}! Skipped')
            continue
        if not res.missing_hydrogens and not rebuild:
            continue
        res_builder.load_residue(res)
        if rebuild:
            res_builder.remove_hydrogens()
        built_atoms[(res.id[1], res.resname)] = res_builder.build_hydrogens()
    return built_atoms

def fix_chain(chain):
    """Fix a PolymerChain by building missing atoms and hydrogens based on 
    topology definitions. Missing atom will be built based on the IC definitions 
    if the residue is not completely missing (e.g. missing loop on a chain). Any 
    undefined atoms will be removed.
    
    Args:
        chain: PolymerChain with topology definitions loaded for each residue.
    
    Returns:
        A dictionary of built atoms with residue id and residue name as keys.
    """
    built_atoms = {}
    res_builder = ResidueFixer()
    chain.sort_residues()
    for res in chain:
        if res.topo_definition is None:
            warnings.warn(f'No topology definition on {res}! Skipped')
            continue
        if res.missing_atoms or res.missing_hydrogens:
            res_builder.load_residue(res)
            cur_built_atoms = built_atoms[(res.id[1], res.resname)] = []
            if built_heavy_atoms := res_builder.build_missing_atoms():
                cur_built_atoms.extend(built_heavy_atoms)
            # we force rebuild hydrogens if there are missing heavy atoms
            built_hydrogens = res_builder.rebuild_hydrogens()
            cur_built_atoms.extend(built_hydrogens)
        if res.undefined_atoms:
            res_builder.load_residue(res)
            res_builder.remove_undefined_atoms()
    return built_atoms

def build_water_hs(water):
    """Build missing hydrogens for a water molecule. The water molecule must have
    a topology definition loaded.
    
    Args:
        water: Water molecule with topology definition loaded.
    
    Returns:
        A list of built hydrogen atoms.
    """
    if water.topo_definition is None:
        warnings.warn(f'No topology definition on {water}! Skipped')
        return []
    res_builder = ResidueFixer()
    res_builder.load_residue(water)
    return res_builder.build_hydrogens()